@@ -176,7 +176,7 @@ class CrashReportDatabaseGeneric : public CrashReportDatabase {
   // CrashReportDatabase:
   Settings* GetSettings() override;
   OperationStatus PrepareNewCrashReport(
-      std::unique_ptr<NewReport>* report) override;
+      std::unique_ptr<NewReport>* report, const UUID* uuid) override;
   OperationStatus FinishedWritingCrashReport(std::unique_ptr<NewReport> report,
                                              UUID* uuid) override;
   OperationStatus LookUpCrashReport(const UUID& uuid, Report* report) override;
@@ -332,12 +332,15 @@ Settings* CrashReportDatabaseGeneric::GetSettings() {
 }
 
 OperationStatus CrashReportDatabaseGeneric::PrepareNewCrashReport(
-    std::unique_ptr<NewReport>* report) {
+    std::unique_ptr<NewReport>* report, const UUID* uuid) {
   INITIALIZATION_STATE_DCHECK_VALID(initialized_);
 
   auto new_report = std::make_unique<NewReport>();
   if (!new_report->Initialize(
-          this, base_dir_.Append(kNewDirectory), kCrashReportExtension)) {
+          this,
+          base_dir_.Append(kNewDirectory),
+          kCrashReportExtension,
+          uuid)) {
     return kFileSystemError;
   }
 
