@@ -181,6 +181,18 @@ sentry__envelope_set_header(
 
 sentry_envelope_t *
 sentry__envelope_new(void)
+{
+    sentry_dsn_t *dsn = NULL;
+    SENTRY_WITH_OPTIONS (options) {
+        dsn = sentry__dsn_incref(options->dsn);
+    }
+    sentry_envelope_t *rv = sentry__envelope_new_with_dsn(dsn);
+    sentry__dsn_decref(dsn);
+    return rv;
+}
+
+sentry_envelope_t *
+sentry__envelope_new_with_dsn(const sentry_dsn_t *dsn)
 {
     sentry_envelope_t *rv = SENTRY_MAKE(sentry_envelope_t);
     if (!rv) {
@@ -193,11 +205,9 @@ sentry__envelope_new(void)
     rv->contents.items.item_count = 0;
     rv->contents.items.headers = sentry_value_new_object();
 
-    SENTRY_WITH_OPTIONS (options) {
-        if (options->dsn && options->dsn->is_valid) {
-            sentry__envelope_set_header(rv, "dsn",
-                sentry_value_new_string(sentry_options_get_dsn(options)));
-        }
+    if (dsn && dsn->is_valid) {
+        sentry__envelope_set_header(
+            rv, "dsn", sentry_value_new_string(dsn->raw));
     }
 
     return rv;
