@@ -458,9 +458,9 @@ sentry_log_fatal(const char *message, ...)
 }
 
 void
-sentry__logs_startup(void)
+sentry__logs_startup(const sentry_options_t *options)
 {
-    sentry__batcher_startup(&g_batcher);
+    sentry__batcher_startup(&g_batcher, options);
 }
 
 bool
